#include "MY_DS3231.h"

#include "stdio.h"
#include "string.h"

#include "esp_types.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

uint8_t dec2Bcd(uint8_t val) {

  return (((val / 10) << 4) + (val % 10));
};

uint8_t bcd2Dec(uint8_t val) {

  return ((val & 0x0F) + (val >> 4) * 10);
};

int initRTC(i2c_dev_t *RTCStruct) {

  //   i2c_config_t i2c_config = {

  //       .mode = I2C_MODE_MASTER,
  //       .sda_io_num = RTCStruct->SDAPin,
  //       .scl_io_num = RTCStruct->SCLPin,
  //       .sda_pullup_en = GPIO_PULLUP_DISABLE,
  //       .scl_pullup_en = GPIO_PULLUP_DISABLE,
  //       .master.clk_speed = 100000};

  i2cdev_init();

  i2c_dev_create_mutex(RTCStruct);

  return (1);
};

int getRawRTCTime(i2c_dev_t *RTCStruct, void *Output[]) {

  uint8_t RawRegister[7];

  memset(&RawRegister, 0, 7 * sizeof(uint8_t));

  // i2c_cmd_handle_t cmd_handle = i2c_cmd_link_create();
  // i2c_master_start(cmd_handle);

  // // Request a register read request from device by sending it'd address ORed with a READ BIT
  // i2c_master_write_byte(cmd_handle, (RTCStruct->device_addr << 1) | I2C_MASTER_READ, true);
  // i2c_master_read(cmd_handle, (uint8_t *) &RawRegister, 18, I2C_MASTER_ACK);
  // i2c_master_read(cmd_handle, (uint8_t *) &RawRegister[18], 1, I2C_MASTER_NACK);
  // i2c_master_stop(cmd_handle);

  // // This tells you when it ends?
  // i2c_master_cmd_begin(I2C_NUM_0, cmd_handle, 1000/portTICK_RATE_MS);

  // i2c_cmd_link_delete(cmd_handle); // removes the I2C handler to free memory for the system

  I2C_DEV_TAKE_MUTEX(RTCStruct);
  I2C_DEV_CHECK(RTCStruct, i2c_dev_read_reg(RTCStruct, 0, Output, 7));
  I2C_DEV_GIVE_MUTEX(RTCStruct);

  //   for (int i = 0; i < sizeof(RawRegister); i++) {

  //     Output[i] = RawRegister[i];
  //   }

  return (1);
};

int getRTCTime(i2c_dev_t *RTCStruct, struct tm *TimeO) {

  uint8_t RawRegister[7];

  I2C_DEV_TAKE_MUTEX(RTCStruct);
  I2C_DEV_CHECK(RTCStruct, i2c_dev_read_reg(RTCStruct, 0, &RawRegister, 7));
  I2C_DEV_GIVE_MUTEX(RTCStruct);

  TimeO->tm_sec = bcd2Dec(RawRegister[0]);
  TimeO->tm_min = bcd2Dec(RawRegister[1]);
  TimeO->tm_hour = bcd2Dec(RawRegister[2]);
  TimeO->tm_wday = bcd2Dec(RawRegister[3]) + 1;
  TimeO->tm_mday = bcd2Dec(RawRegister[4]);
  TimeO->tm_mon = bcd2Dec(RawRegister[5]) + 1;
  TimeO->tm_year = bcd2Dec(RawRegister[6]) + 100;

  return (1);
};

int setRTCTime(i2c_dev_t *RTCStruct, struct tm *TimeI) {

  uint8_t RawRegister[7];

  RawRegister[0] = dec2Bcd(TimeI->tm_sec);
  RawRegister[1] = dec2Bcd(TimeI->tm_min);
  RawRegister[2] = dec2Bcd(TimeI->tm_hour);
  RawRegister[3] = dec2Bcd(TimeI->tm_wday - 1);
  RawRegister[4] = dec2Bcd(TimeI->tm_mday);
  RawRegister[5] = dec2Bcd(TimeI->tm_mon - 1);
  RawRegister[6] = dec2Bcd(TimeI->tm_year - 100);

  I2C_DEV_TAKE_MUTEX(RTCStruct);
  I2C_DEV_CHECK(RTCStruct, i2c_dev_write_reg(RTCStruct, 0, &RawRegister, 7));
  I2C_DEV_GIVE_MUTEX(RTCStruct);

  return (1);
};

int setRTCClock(i2c_dev_t *RTCStruct, uint8_t Sec, uint8_t Minute, uint8_t Hour, uint8_t Day,
             uint8_t Date, uint8_t Month, uint8_t Year) {

  uint8_t RawRegister[7];

  RawRegister[0] = dec2Bcd(Sec);
  RawRegister[1] = dec2Bcd(Minute);
  RawRegister[2] = dec2Bcd(Hour);
  RawRegister[3] = dec2Bcd(Day - 1);
  RawRegister[4] = dec2Bcd(Date);
  RawRegister[5] = dec2Bcd(Month - 1);
  RawRegister[6] = dec2Bcd(Year - 100);

  I2C_DEV_TAKE_MUTEX(RTCStruct);
  I2C_DEV_CHECK(RTCStruct, i2c_dev_write_reg(RTCStruct, 0, RawRegister, 7));
  I2C_DEV_GIVE_MUTEX(RTCStruct);

  return (1);
};