#include "button.h"
#include "stdio.h"

#include "driver/gpio.h"

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"

#include "esp_err.h"

#include "esp_timer.h"


xQueueHandle btnQueue;

// // Dont printf() from ISR !!!!!!!!!!!!!

static void IRAM_ATTR buttonISR(void *args){

  btn_params_t *BtnParams = (btn_params_t*)&args;
  xQueueSendFromISR(btnQueue, BtnParams, NULL);

}

void btnTask(void *pvParams){

  int pinNumber, count = 0;
  btn_params_t* ParamsPointer;
  btn_params_t BtnParams;

  // Debounce time
 

   

  while (true)
  {
    
    if(xQueueReceive(btnQueue, &ParamsPointer, portMAX_DELAY)){
      
      BtnParams = *ParamsPointer;

      // printf("State change: %d @ PinNum: %d \n", gpio_get_level(BtnParams.BtnPin), BtnParams.BtnPin);
      gpio_isr_handler_remove(BtnParams.BtnPin);
      // vTaskDelay(pdMS_TO_TICKS(1));
      // printf("Button %d was pressed %d times The state is %d\n", pinNumber, count++, gpio_get_level(pinNumber));

      ParamsPointer->CS = gpio_get_level(ParamsPointer->BtnPin);

      if(ParamsPointer->CS == 0 && ParamsPointer->LS == 1){
        
        // Do work for Pressed
        if(ParamsPointer->PressedISR != NULL && ParamsPointer->AF == 0){
          ParamsPointer->AF = 1;
          ParamsPointer->PressedISR(ParamsPointer->arg);
        }
        
        ParamsPointer->CurrTime = (esp_timer_get_time()/1000);
        
        ParamsPointer->LS = ParamsPointer->CS;
      }
      else{
        
        ParamsPointer->DBT = (esp_timer_get_time()/1000);

        if((ParamsPointer->DBT - ParamsPointer->CurrTime) > 20 && ParamsPointer->AF){
          
          // This code executes after button release
          printf("Do work # %d \n", count++);
          // myFunc();

          if(ParamsPointer->ReleasedISR != NULL){
            ParamsPointer->ReleasedISR(ParamsPointer->arg);
          }
          ParamsPointer->AF = 0;
        }
        ParamsPointer->LS = ParamsPointer->CS;
        
      }
      gpio_isr_handler_add(ParamsPointer->BtnPin, buttonISR, ParamsPointer);
    }
  }

}


uint8_t initButtons(btn_params_t **BtnsPointer, uint8_t ButtonCount){

      // Initialize input pin 
//   gpio_set_direction(BtnParams->BtnPin, GPIO_MODE_INPUT);
//   gpio_pulldown_dis(BtnParams->BtnPin);
//   gpio_pullup_en(BtnParams->BtnPin);

  printf("%d \n", BtnsPointer[2]->BtnPin);
  gpio_config_t BtnConfig;

//   BtnParams->BtnActiveState;

  BtnConfig.intr_type = GPIO_INTR_ANYEDGE;
  BtnConfig.mode = GPIO_MODE_INPUT;
  BtnConfig.pull_down_en = BtnsPointer[0]->BtnPinPullDown;
  BtnConfig.pull_up_en = BtnsPointer[0]->BtnPinPullUp;

  BtnConfig.pin_bit_mask = 0;
  printf("Size of BtnsPins %d \n", sizeof(&BtnsPointer));

  for(short I = 0; I < ButtonCount; I++){
    BtnsPointer[I]->DBT = 0;
    BtnsPointer[I]->CurrTime = 0;
    BtnsPointer[I]->AF = 0;
    BtnsPointer[I]->LS = 1;
    BtnsPointer[I]->CS = 1;

    BtnsPointer[I]->PressedISR = NULL;
    BtnsPointer[I]->ReleasedISR = NULL;

    BtnConfig.pin_bit_mask |= (1ULL << BtnsPointer[I]->BtnPin);
  }

  gpio_config(&BtnConfig);


// gpio_set_intr_type(BtnsParams->BtnPin, GPIO_INTR_NEGEDGE);
  gpio_install_isr_service(0);
  



  // printf("Initialized button #%d \n", BtnsParams->BtnPins[0]);

  return(1);
}


uint8_t initButton(btn_params_t *BtnParams){

      // Initialize input pin 
//   gpio_set_direction(BtnParams->BtnPin, GPIO_MODE_INPUT);
//   gpio_pulldown_dis(BtnParams->BtnPin);
//   gpio_pullup_en(BtnParams->BtnPin);

  gpio_config_t BtnConfig;

//   BtnParams->BtnActiveState;

  BtnConfig.intr_type = GPIO_INTR_ANYEDGE;
  BtnConfig.mode = GPIO_MODE_INPUT;
  BtnConfig.pull_down_en = BtnParams->BtnPinPullDown;
  BtnConfig.pull_up_en = BtnParams->BtnPinPullUp;
  BtnConfig.pin_bit_mask = (1ULL << BtnParams->BtnPin);

  gpio_config(&BtnConfig);


// gpio_set_intr_type(BtnParams->BtnPin, GPIO_INTR_NEGEDGE);
  gpio_install_isr_service(0);
  



  printf("Initialized button #%d \n", BtnParams->BtnPin);

  return(1);
}

void startButtons(){

  // Create queue to communicate between buttonISR and task
  btnQueue = xQueueCreate(10, sizeof(gpio_config_t));
  // Create the button task ( what to do when button is pressed)
  xTaskCreate(btnTask, "TriggeredBtn", 2048, NULL, 1, NULL);
}


// Make a button group struct with button pin's array, button floating state, button pull, 
// function that takes the button group handler and callback function
// Add attachHold with additional hold time activation
void attachReleased(btn_params_t *BtnParams, void (*BtnISR)(void *args), void *arg){
  // printf("Attached pin #%d\n ", BtnParams->BtnPin);
  BtnParams->ReleasedISR = BtnISR;
  BtnParams->arg = arg;
  // BtnParams->args = args;
  gpio_isr_handler_add(BtnParams->BtnPin, buttonISR, (btn_params_t *)BtnParams);
}

void attachReleased_s(btns_params_t *BtnsParams, uint8_t Idx, void (*BtnISR)()){
  // printf("Attached pin #%d\n ", BtnParams->BtnPin);
  BtnsParams->ReleasedISR[Idx] = BtnISR;
  BtnsParams->ISRIdx = Idx;
  gpio_isr_handler_add(BtnsParams->BtnsPins[BtnsParams->ISRIdx], buttonISR, (void *)BtnsParams);
}

void attachPressed(btn_params_t *BtnParams, void (*BtnISR)(void *args), void *arg){

  BtnParams->PressedISR = BtnISR;
  BtnParams->arg = arg;
  // BtnParams->args = args;
  gpio_isr_handler_add(BtnParams->BtnPin, buttonISR, (btn_params_t *)BtnParams);
  
}