#include "my_buzzer.h"
// Buzzer driver lib (PWM)
#include "driver/ledc.h"

#include "esp_err.h"

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

esp_err_t buzzer_init(uint8_t BuzzerPin, ledc_channel_t _channel_num, ledc_timer_t _timer_num, float Freq) {

  ledc_timer_config_t timer = {
      .speed_mode = GPIO_OUTPUT_SPEED,
      .duty_resolution = LEDC_TIMER_10_BIT,
      .timer_num = _timer_num,
      .freq_hz = Freq,
      .clk_cfg = LEDC_AUTO_CLK};

  ledc_timer_config(&timer);

  ledc_channel_config_t channel = {
      .gpio_num = BuzzerPin,
      .speed_mode = GPIO_OUTPUT_SPEED,
      .channel = _channel_num,
      .timer_sel = _timer_num,
      .duty = 0,
      .hpoint = 0};
  ledc_channel_config(&channel);

  return (ESP_OK);
}

void buzzer_tone(float Freq) {

  ledc_timer_config_t timer = {
      .speed_mode = GPIO_OUTPUT_SPEED,
      .duty_resolution = LEDC_TIMER_10_BIT,
      .timer_num = LEDC_TIMER_0,
      .freq_hz = Freq,
      .clk_cfg = LEDC_AUTO_CLK};

  ledc_timer_config(&timer);

  // ledc_set_duty_and_update(GPIO_OUTPUT_SPEED, LEDC_CHANNEL_0, 512, 0);
  ledc_set_freq(GPIO_OUTPUT_SPEED, LEDC_CHANNEL_0, Freq);
  ledc_set_duty(GPIO_OUTPUT_SPEED, LEDC_CHANNEL_0, 512); // 12% duty - play here for your speaker or buzzer
  ledc_update_duty(GPIO_OUTPUT_SPEED, LEDC_CHANNEL_0);
}

void buzzer_stop() {

  ledc_set_duty(GPIO_OUTPUT_SPEED, LEDC_CHANNEL_0, 0);
  ledc_update_duty(GPIO_OUTPUT_SPEED, LEDC_CHANNEL_0);
}

void buzzer_beep(uint BeepCount, float Freq, int OnPeriod, uint OffPeriod) {

  for (uint8_t C = 0; C < BeepCount; C++) {

    buzzer_tone(Freq);
    vTaskDelay(OnPeriod / portTICK_PERIOD_MS);
    buzzer_stop();
    vTaskDelay(OffPeriod / portTICK_PERIOD_MS);
  }
}