#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"
#include "freertos/semphr.h"
#include "freertos/task.h"

#include "driver/gpio.h"

#include "esp_err.h"
#include "esp_log.h"

#include "button.h"
#include "my_buzzer.h"


#include <u8g2.h>
#include <u8g2_esp32_hal.h>

#include "MY_DS3231.h"

#include "esp_sntp.h"
#include "esp_wifi.h"

#include "esp_http_client.h"
#include "protocol_examples_common.h"

#include "esp_event.h"
#include "nvs_flash.h"

#include "time.h"

#include "driver/timer.h"
#include "esp_timer.h"

#define MAX_APs 20

#define Lcdheight 64
#define Lcdwidth 128
// #define VbiasPOT 140 // contrast 00 to FF , default 0x49 , user adjust

#define I2C_SDA_PIN 21
#define I2C_SCL_PIN 22

// GPIO 5-wire SPI interface


#define BTN0_PIN 32 // Toggle time
#define BTN1_PIN 33 // Record time
#define BTN2_PIN 25 // Reset time

#define BTN3_PIN 26 // Up strok
#define BTN4_PIN 27 // Mid strok
#define BTN5_PIN 14 // Down Strok

#define BUZZER_PIN 12

#define RTC_ADDR 0x68

#define MAIN_TIMER_GRP = TIMER_GROUP_0
#define MAIN_TIMER_IDX = TIMER_0
#define TIMER_DIVIDER 80 // At 80MhZ base clock it means 1uS per finale tick

// #define WIFI_NTP_SYNC 1
int8_t Page;
int16_t Cursor;

static volatile int64_t RefTimePoint;
static volatile int64_t PauseTimePoint = 0;
static volatile int64_t FinaleTimePoint = 0;

bool TimerStatus = 0;

xQueueHandle RTCTimeQ;
xQueueHandle NTPReadyQ;

xQueueHandle TimerTickQ;
xQueueHandle ClockTickQ;

xQueueHandle TimerRecordQ;

xQueueHandle MenuQ;

typedef struct {

  int8_t Page;
  int8_t Cursor;
} menu_map_t;

xSemaphoreHandle CurrTimerTimeS;

// xQueueHandle TimerQ;
static SemaphoreHandle_t TimerS;

esp_timer_handle_t DispTimerHandle;
esp_timer_handle_t ClockTHandle;

typedef struct {

  uint64_t RecordTime;
  int64_t SinceBootTime;
  time_t RTCTime;

} timer_record_t;

timer_record_t RecordedTimes[100];

short TimerIndex = 0;

int64_t CurrTimerTime;

nvs_handle NVSHandle;

void toggleTimer() {

  uint64_t CTimer;
  timer_config_t TimerConf;

  timer_get_counter_value(TIMER_GROUP_0, TIMER_0, &CTimer);
  timer_get_config(TIMER_GROUP_0, TIMER_0, &TimerConf);
  // if timer is enabled disable else enable(start)
  if (TimerConf.counter_en) {

    timer_pause(TIMER_GROUP_0, TIMER_0);
    esp_timer_stop(DispTimerHandle);
    // PauseTimePoint += CTimer - RefTimePoint;
    // RefTimePoint = CTimer;

    printf("Stopping the timer, %lld\n", CTimer);
  }

  else {
    timer_start(TIMER_GROUP_0, TIMER_0);
    esp_timer_start_periodic(DispTimerHandle, 100000);
    // RefTimePoint = CTimer;
    TimerStatus = 1;
    printf("Starting the timer, %lld\n", CTimer);
  }
  xQueueSend(TimerTickQ, &CTimer, pdMS_TO_TICKS(1000));
}

void recordTimer() {

  // static int64_t CurrTimerTime;

  // CurrTimerTime = (esp_timer_get_time() - RefTimePoint) + PauseTimePoint;
  // xSemaphoreTake(CurrTimerTimeS, 1000/portTICK_PERIOD_MS);
  timer_get_counter_value(TIMER_GROUP_0, TIMER_0, &RecordedTimes[TimerIndex].RecordTime);
  // xSemaphoreGive(CurrTimerTimeS);

  RecordedTimes[TimerIndex].SinceBootTime = esp_timer_get_time();
  time(&RecordedTimes[TimerIndex].RTCTime);

  TimerIndex++;
  xQueueSend(TimerRecordQ, &RecordedTimes[TimerIndex], pdMS_TO_TICKS(1000));

  printf("Recorded Times\n");

  // ESP_ERROR_CHECK(nvs_open_from_partition("Myrecords", "TRecords", NVS_READWRITE, &NVSHandle));

  // char IdxStr[14];
  // // itoa(TimerIndex, IdxStr, 10);
  // IdxStr[0] = (48 + TimerIndex);
  // IdxStr[1] = NULL;
  // nvs_set_blob(NVSHandle, IdxStr, (void *)&RecordedTimes[TimerIndex], sizeof(timer_record_t));
  // ESP_ERROR_CHECK(nvs_commit(NVSHandle));
  // nvs_close(NVSHandle);
}

void resetTimer() {

  static uint64_t CurrTimerUs;

  // RefTimePoint = esp_timer_get_time();
  // PauseTimePoint = 0;
  timer_set_counter_value(TIMER_GROUP_0, TIMER_0, 0);
  timer_get_counter_value(TIMER_GROUP_0, TIMER_0, &CurrTimerUs);
  // xSemaphoreTake(CurrTimerTimeS, 1000/portTICK_PERIOD_MS);
  // CurrTimerTime = 0;
  // CurrTimerUs = RefTimePoint + PauseTimePoint;
  xQueueSend(TimerTickQ, &CurrTimerTime, pdMS_TO_TICKS(1000));
  // xSemaphoreGive(CurrTimerTimeS);

  printf("Timer Reset!\n");
}

void upCursor() {
}
void midCursor() {
}
void downCursor() {
}


esp_err_t clientEvent(esp_http_client_event_t *evt) {

  switch (evt->event_id) {

  case HTTP_EVENT_ON_DATA:
    printf("%.*s\n", evt->data_len, (char *)evt->data);
    break;

  default:
    break;
  }

  return ESP_OK;
}


void on_got_time(struct timeval *tv) {

  // printf("GotTime\n");
  xQueueSend(NTPReadyQ, tv, pdMS_TO_TICKS(2000));
}

static bool IRAM_ATTR onDispTimerTick(void *args) {

  BaseType_t high_task_awoken = false;

  static uint64_t CurrTimerUs;

  // xSemaphoreTakeFromISR(CurrTimerTimeS, &high_task_awoken);
  timer_get_counter_value(TIMER_GROUP_0, TIMER_0, &CurrTimerUs);

  // CurrTimerTime = CurrTimerUs;
  // xSemaphoreGiveFromISR(CurrTimerTimeS, &high_task_awoken);
  xQueueSendFromISR(TimerTickQ, &CurrTimerUs, &high_task_awoken);

  return high_task_awoken == true;
}

static bool IRAM_ATTR onClockTick(void *args) {

  BaseType_t high_task_awoken = false;
  // xSemaphoreGiveFromISR(TimerS, &high_task_awoken);
  static time_t CurrClockTime;
  time(&CurrClockTime);
  xQueueSendFromISR(ClockTickQ, &CurrClockTime, &high_task_awoken);

  return high_task_awoken == true;
}

void mainRoutine(void *pvParams) {
  
  timer_config_t TimerConfig = {
    .divider = TIMER_DIVIDER,
    .counter_dir = TIMER_COUNT_UP,
    .counter_en = TIMER_PAUSE,
    .alarm_en = TIMER_ALARM_EN,
    .auto_reload = TIMER_AUTORELOAD_DIS

  };

  i2c_dev_t MyRTC = {

    .port = I2C_NUM_0,
    .addr = RTC_ADDR,
    .cfg.sda_io_num = I2C_SDA_PIN,
    .cfg.scl_io_num = I2C_SCL_PIN,
    .cfg.master.clk_speed = 400000,

  };

  // ---------------------- Init Buttons -------------------------
  // size_t length = sizeof(BtnGroup) / sizeof(BtnGroup[0]);


  btn_params_t Btn0Params = {
      .BtnPin = BTN0_PIN,
      .BtnActiveState = 0,
      .BtnPinPullUp = 1,
      .BtnPinPullDown = 0

    };
  btn_params_t Btn1Params = {.BtnPin = BTN1_PIN};
  btn_params_t Btn2Params = {.BtnPin = BTN2_PIN};

  btn_params_t Btn3Params = {.BtnPin = BTN3_PIN};
  btn_params_t Btn4Params = {.BtnPin = BTN4_PIN};
  btn_params_t Btn5Params = {.BtnPin = BTN5_PIN};

  btn_params_t *BtnGroup[6] = {&Btn0Params, &Btn1Params, &Btn2Params, &Btn3Params, &Btn4Params, &Btn5Params};

  initButtons(BtnGroup, 6);
  // initHardware();

  attachPressed(&Btn0Params, toggleTimer, NULL);
  attachPressed(&Btn1Params, recordTimer, NULL);
  attachPressed(&Btn2Params, resetTimer, NULL);

  attachPressed(&Btn3Params, upCursor, NULL);
  attachPressed(&Btn4Params, midCursor, NULL);
  attachPressed(&Btn5Params, downCursor, NULL);

  startButtons();

  struct tm *Time = malloc(sizeof *Time);
  struct tm *NTPTime = malloc(sizeof *Time);

  char Buffer[26];

  
  struct timeval RecT;
  time_t RTCTSecs;
  time_t now = 0;

    // ---------------------- Init Timers -------------------------
  timer_init(TIMER_GROUP_0, TIMER_0, &TimerConfig);
  timer_set_counter_value(TIMER_GROUP_0, TIMER_0, 0);

    // for alarm setting you need to take TIMER_BASE_CLK and Timer DIVIDER into considiration
  // The Formula is DelayPerTick =  ((1/TIMER_BASE_CLK)* Divider)
  // TimeToSet = TimeToWait / DelayPerTick

  // @ Current setting 80MHz base clock 16 divider the counter will do 5M counts per Second
  // That means if we set the alarm to 2.5M we will get an alarm every 0.5 Seconds
  // timer_set_alarm_value(TIMER_GROUP_0, TIMER_0, (0.5 *(TIMER_BASE_CLK/TIMER_DIVIDER))); //  now 0.2uSec per tick  (TIMER_BASE_CLK/TIMER_DIVIDER) Gives 1sec delay
  // timer_enable_intr(TIMER_GROUP_0, TIMER_0);
  // timer_isr_callback_add(TIMER_GROUP_0, TIMER_0, onTimerTick, NULL, 0);

  const esp_timer_create_args_t DispTArgs = {
    .callback = &onDispTimerTick,
    .name = "DispTimerTicker"};

  const esp_timer_create_args_t ClockTArgs = {
      .callback = &onClockTick,
      .name = "ClockTicker"};

  ESP_ERROR_CHECK(esp_timer_create(&DispTArgs, &DispTimerHandle));
  ESP_ERROR_CHECK(esp_timer_create(&ClockTArgs, &ClockTHandle));

  ESP_ERROR_CHECK(esp_timer_start_periodic(ClockTHandle, 100000));
  
  // ---------------------- Init DS3231 -------------------------
  initRTC(&MyRTC);

  // ---------------------- Init NVS -------------------------
  static const char *TAG = "NVS";

  ESP_ERROR_CHECK(nvs_flash_init_partition("Myrecords"));
  ESP_ERROR_CHECK(nvs_open_from_partition("Myrecords", "TRecords", NVS_READWRITE, &NVSHandle));

  nvs_stats_t NVSStats;

  nvs_get_stats("Myrecords", &NVSStats);

  ESP_LOGI(TAG, "used: %d, free: %d, total: %d, namespace count: %d", NVSStats.used_entries,
           NVSStats.free_entries, NVSStats.total_entries, NVSStats.namespace_count);

  nvs_close(NVSHandle);

  // ---------------------- Init Buzzer -------------------------
  buzzer_init(BUZZER_PIN, LEDC_CHANNEL_0, LEDC_TIMER_0, 500);

  // Max frequancy according ti TIMER_RESOLUTION is: MaxFreq = 40MHz / (2^BIT_RES) ; // BIT_RES = 10 BIT

  // play buzzer_tone @500Hz
  // buzzer_tone(500);

  // buzzer_stop();
 

  // ---------------------- Init WiFi/SyncClocks -------------------------

  time_t NTPTSecs;

  #ifdef WIFI_NTP_SYNC
  // Init Wifi
  nvs_flash_init();
  tcpip_adapter_init();
  esp_event_loop_create_default();

  example_connect();

  // Init NTP service
  sntp_set_sync_mode(SNTP_SYNC_MODE_IMMED);
  sntp_setoperatingmode(SNTP_OPMODE_POLL);
  sntp_setservername(0, "pool.ntp.org");
  sntp_init();

  struct timeval CurrTime;
  sntp_set_time_sync_notification_cb(on_got_time);
  // sntp_sync_time(&CurrTime); // what does this function do?

  // Wait for NTP to sync with ESP

  if (xQueueReceive(NTPReadyQ, &RecT, pdMS_TO_TICKS(12000))) {
    time(&NTPTSecs);

  } else {
    printf("Failed to receive from Q\n");
  }

  // Sync Timers

  time(&NTPTSecs);
  NTPTime = localtime(&NTPTSecs);
  // char Buffer[26];

  // Compare NTP and RTC outputs before syncing:

  strftime(Buffer, 26, "%H:%M:%S %d/%m-%y ", NTPTime);

  printf("NTP Time: %s\n", Buffer);

  getRTCTime(&MyRTC, Time);

  RTCTSecs = mktime(Time); // Convet struct tm to time_t

  strftime(Buffer, 26, "%H:%M:%S %d/%m-%y ", Time);
  // xQueueSend(RTCTimeQ, Time, pd)
  printf("RTC Time: %s\n", Buffer);

  // Check difference
  RTCTSecs += 1;
  double Difference = 0;

  Difference = difftime(RTCTSecs, NTPTSecs);
  printf("Difference = %f\n", Difference);

  if (Difference > 1 || Difference < -1) {

    time(&NTPTSecs);
    NTPTime = localtime(&NTPTSecs);

    setRTCTime(&MyRTC, NTPTime);
    printf("Set the clock\n");

    getRTCTime(&MyRTC, Time);
    time(&NTPTSecs);
    NTPTime = localtime(&NTPTSecs);

    RTCTSecs = mktime(Time); // Convet struct tm to time_t

    strftime(Buffer, 26, "%H:%M:%S %d/%m-%y ", NTPTime);
    printf("NTP Time: %s\n", Buffer);

    strftime(Buffer, 26, "%H:%M:%S %d/%m-%y ", Time);
    printf("RTC Time: %s\n", Buffer);

    printf("Difference = %f\n", difftime(RTCTSecs, NTPTSecs));
  }
#endif

  
#ifndef WIFI_NTP_SYNC

  getRTCTime(&MyRTC, Time);

  struct timeval *tv = malloc(sizeof(struct timeval));
  tv->tv_sec = mktime(Time);

  setenv("TZ", "UTC-3", 1);
  tzset();

  settimeofday(tv, NULL);

#endif


  while (true) {
    // getRTCTime(&MyRTC, Time);
    // strftime(Buffer, 26, "%H:%M:%S %d/%m-%y ", Time);
    // printf("RTC Time: %s\n", Buffer);

    vTaskDelay(pdMS_TO_TICKS(1000));
  }
}

void displayRoutine(void *pvParams) {

  vTaskDelay(pdMS_TO_TICKS(10));

  static const char *TAG = "Display";
  printf("Start print to display\n");

  gpio_config_t BLPinConfig;

  BLPinConfig.intr_type = GPIO_INTR_DISABLE;
  BLPinConfig.mode = GPIO_MODE_OUTPUT;
  BLPinConfig.pull_down_en = 0;
  BLPinConfig.pull_up_en = 0;
  BLPinConfig.pin_bit_mask = ((1ULL << DISP_BL_PIN)); // must be bit shifted left (1ULL <<)

  // Init input pins
  gpio_config(&BLPinConfig);

  gpio_set_level(DISP_BL_PIN, 1);

  u8g2_esp32_hal_t u8g2_esp32_hal = U8G2_ESP32_HAL_DEFAULT;
  u8g2_esp32_hal.mosi = DISP_MOSI;
  u8g2_esp32_hal.clk = DISP_CLK;
  u8g2_esp32_hal.cs = DISP_CS;
  u8g2_esp32_hal.dc = DISP_DC;
  u8g2_esp32_hal.reset = DISP_RST;

  // Init display
  u8g2_esp32_hal_init(u8g2_esp32_hal);

  u8g2_t u8g2; // a structure which will contain all the data for one display

  u8g2_Setup_st7567_enh_dg128064_f(
      &u8g2,
      U8G2_R2, // Display Rotation U8G2_R0-NoRot, U8G2_R1-90Deg U8G2_R2-180Deg
      u8g2_esp32_spi_byte_cb,
      u8g2_esp32_gpio_and_delay_cb); // init u8g2 structure

  ESP_LOGI(TAG, "u8g2_InitDisplay");
  u8g2_InitDisplay(&u8g2); // send init sequence to the display, display is in sleep mode after this,

  ESP_LOGI(TAG, "u8g2_SetPowerSave");
  u8g2_SetPowerSave(&u8g2, 0); // wake up display

  u8g2_SetContrast(&u8g2, 220);

  u8g2_SetFontPosTop(&u8g2);

  u8g2_ClearBuffer(&u8g2); // clear the internal memory
  u8g2_SendBuffer(&u8g2);

  u8g2_SetFont(&u8g2, u8g2_font_7x14B_tr);
  u8g2_DrawUTF8(&u8g2, 20, 20, "MY Timer PRJ");

  u8g2_SendBuffer(&u8g2); // transfer internal memory to the display

  int FontH, FontW;
  u8g2_SetFont(&u8g2, u8g2_font_lubB14_tn); // set the target font for the text width calculation

  FontH = u8g2_GetMaxCharHeight(&u8g2); // Also works with Width
  FontW = u8g2_GetMaxCharWidth(&u8g2);

  printf("Font size W: %d, H: %d\n", FontW, FontH);

  vTaskDelay(pdMS_TO_TICKS(1000));

  char Buff[32];
  char TimeBuff[24];

  static int64_t CurrTimer;
  static int64_t CurrRecord;
  time_t CurrClock;

  struct tm *TimerDisp = malloc(sizeof(*TimerDisp));
  struct tm *RecTimerDisp = malloc(sizeof(*RecTimerDisp));

  struct tm *ClockDisp = malloc(sizeof(*ClockDisp));

  struct timeval *tv;
  // time_t Time;
  long TimerD;
  int64_t TBuff;

  u8g2_ClearBuffer(&u8g2);

  // Force initial display on data:
  onClockTick(NULL);
  onDispTimerTick(NULL);

  menu_map_t MenuMap;

  // int8_t Page = 0, Cursor = 0;

  while (1) {

    if (xQueueReceive(MenuQ, &MenuMap, 1)) {

      switch (Page){
        case 0:
          MenuMap.Cursor = TimerIndex;
        break;
      }
      // Change brush to xor
      u8g2_SetDrawColor(&u8g2, 2);
      // Mark reading according to cursor index
      u8g2_DrawFrame(&u8g2, 0, 27, 8, 10);
    }

    if (xQueueReceive(ClockTickQ, &CurrClock, 1)) {

      ClockDisp = localtime(&CurrClock); // Convert CurrClock to struct tm ClockDisp
      clearLine(&u8g2, 0, 0, 6, 8, 20);

      strftime(Buff, 26, "%H:%M:%S %d/%m/%y", ClockDisp);
      u8g2_SetFont(&u8g2, u8g2_font_ncenB08_tr); // choose a suitable font u8g2_font_ncenB08_tr
      u8g2_DrawStr(&u8g2, 0, 0, Buff);           // write something to the internal memory
    }

    if (xQueueReceive(TimerTickQ, &CurrTimer, 1)) {

      // time(&CurrClock); // Read Current time into CurrClock
      // printf("%lld, Ref = %lld\n", esp_timer_get_time(), RefTimePoint);

      // CurrTimer = ((esp_timer_get_time() - RefTimePoint) + PauseTimePoint)*TimerStatus;

      CurrTimer *= TimerStatus; // Check if timer display in on to prevent unindendent unproccessed nuymber display

      // xSemaphoreTake(CurrTimerTimeS, 1000/portTICK_PERIOD_MS);

      TBuff = ((CurrTimer / 1000000) / 3600);
      TimerDisp->tm_hour = TBuff;

      TimerDisp->tm_min = ((CurrTimer / 1000000) / 60 % 60);
      TimerDisp->tm_sec = (CurrTimer / (1000000)) % 60;
      // TimerDisp->tm_msec = CurrTimer/(1000);
      TimerDisp->tm_mday = (CurrTimer / (1000) % 1000);

      strftime(Buff, 26, "%H:%M:%S:%d", TimerDisp);
      // sprintf(Buff, "%d:%d:%d", TimerD);
      clearLine(&u8g2, 0, 10, 7, 12, 18);

      u8g2_SetFont(&u8g2, u8g2_font_lubB14_tn); // choose a suitable font u8g2_font_ncenB08_tr
      u8g2_DrawStr(&u8g2, 0, 10, Buff);         // write something to the internal memory

      u8g2_SendBuffer(&u8g2);
    }

    if (xQueueReceive(TimerRecordQ, &CurrRecord, 1)) {

      for (int8_t I = 1; I <= 3; I++) {

        // Check if there is still TimerRecords to to display from this loop to 3
        if (TimerIndex - I < 0)
          break;

        TBuff = ((RecordedTimes[TimerIndex - I].RecordTime / 1000000) / 3600);
        RecTimerDisp->tm_hour = TBuff;

        RecTimerDisp->tm_min = ((RecordedTimes[TimerIndex - I].RecordTime / 1000000) / 60 % 60);
        RecTimerDisp->tm_sec = (RecordedTimes[TimerIndex - I].RecordTime / (1000000)) % 60;
        // TimerDisp->tm_msec = RecordedTimes[(TimerIndex-I)].RecordTime/(1000);
        RecTimerDisp->tm_mday = (RecordedTimes[TimerIndex - I].RecordTime / (1000) % 1000);

        strftime(TimeBuff, 28, "%H:%M:%S:%d", RecTimerDisp);

        clearLine(&u8g2, 0, (28 + 9 * (I - 1)), 6, 8, 14);

        sprintf(Buff, "%d %s", ((TimerIndex - I) + 1), TimeBuff);

        u8g2_SetFont(&u8g2, u8g2_font_ncenB08_tr); // choose a suitable font u8g2_font_ncenB08_tr
        u8g2_DrawStr(&u8g2, 0, (28 + 9 * (I - 1)), Buff);
        u8g2_SendBuffer(&u8g2);
      }
    }

    vTaskDelay(1);
  }
}


void app_main(void) {

  RTCTimeQ = xQueueCreate(5, sizeof(struct tm));
  NTPReadyQ = xQueueCreate(5, sizeof(struct timeval));

  TimerTickQ = xQueueCreate(10, sizeof(uint64_t));
  ClockTickQ = xQueueCreate(5, sizeof(time_t));

  MenuQ = xQueueCreate(5, sizeof(int8_t));

  TimerRecordQ = xQueueCreate(10, sizeof(timer_record_t));

  // TimerQ = xQueueCreate(2, sizeof(time_t));
  TimerS = xSemaphoreCreateBinary();

  CurrTimerTimeS = xSemaphoreCreateMutex();
  xTaskCreatePinnedToCore(displayRoutine, "disp_routine", configMINIMAL_STACK_SIZE * 20, NULL, 2, NULL, APP_CPU_NUM);
  xTaskCreatePinnedToCore(mainRoutine, "main_routine", configMINIMAL_STACK_SIZE * 10, NULL, 2, NULL, APP_CPU_NUM);
}
